"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

from .bast import Bast
from .controller import Controller
from .hash import Hash
from .migration import CreateMigration
from .migration import Migration
from .model import Models
from .route import Route

__version__ = "1.0"




"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

import os

from git import Repo

from .bast import __version__
from .migration import CreateMigration, Migration

try:
    from configparser import ConfigParser
except ImportError:
    import ConfigParser
import bcrypt
import shutil
import click
import re

""" Handles the CLI commands and their respective Arguments """


@click.group()
@click.version_option(__version__)
def main():
    config_path = os.path.abspath('.') + "/config/config.ini"
    if not os.path.exists(config_path):
        return

    config = ConfigParser()
    config.read(config_path)

    #   config section
    os.environ['APP_NAME'] = config['CONFIG']['APP_NAME']
    os.environ['APP_KEY'] = config['CONFIG']['APP_KEY']

    os.environ['DB_TYPE'] = config['DATABASE']['DB_TYPE']
    os.environ['DB_NAME'] = config['DATABASE']['DB_NAME']
    os.environ['DB_HOST'] = config['DATABASE']['DB_HOST']
    os.environ['DB_USER'] = config['DATABASE']['DB_USER']
    os.environ['DB_PASSWORD'] = config['DATABASE']['DB_PASSWORD']
    os.environ['DB_PREFIX'] = config['DATABASE']['DB_PREFIX']

    # print(os.environ['APP_NAME'])
    # pass


@main.command('create:controller', short_help='Creates a Controller File')
@click.argument('filename', required=1)
def controller_creatr(filename):
    """Name of the controller file to be created"""
    bast_app = os.path.abspath('.') + '/config/config.ini'
    if not os.path.exists(bast_app):
        click.echo('ERROR: Ensure you are in a bast app to run the create:controller command')
        return

    path = os.path.abspath('.') + '/controller'
    if not os.path.exists(path):
        os.makedirs(path)

    file_name = str(filename + '.py')
    controller_file = open(os.path.abspath('.') + '/controller/' + file_name, 'w+')
    compose = "from bast import Controller\n\nclass " + filename + "(Controller):\n    pass"
    controller_file.write(compose)
    controller_file.close()
    click.echo("\033[1;32;40m Modes/Controller " + filename + " created successfully")


@main.command('create:view', short_help="Create a View File")
@click.argument('filename', required=1)
def view_creatr(filename):
    """Name of the View File to be created"""
    bast_app = os.path.abspath('.') + '/config/config.ini'
    if not os.path.exists(bast_app):
        click.echo('ERROR: Ensure you are in a bast app to run the create:view command')
        return

    path = os.path.abspath('.') + '/public/templates'
    if not os.path.exists(path):
        os.makedirs(path)

    filename_ = str(filename + ".html").lower()
    view_file = open(path + "/" + filename_, 'w+')
    view_file.write("")
    view_file.close()


@main.command('run', short_help="Run your Bast Server")
@click.option('--serverfile', help="Name of the file to run", default='server.py')
def run(serverfile):
    bast_app = os.path.abspath('.') + '/config/config.ini'
    if not os.path.exists(bast_app):
        click.echo('ERROR: Ensure you are in a bast app to use the "run" command')
        return

    cmd = 'python ' + serverfile
    os.system(cmd)


@main.command('new', short_help="Create a new Bast Project")
@click.argument('projectname', required=1)
def create_new(projectname):
    """Name of the project"""
    git_url = "https://github.com/moluwole/Bast_skeleton"
    path = os.path.abspath('.') + "/" + projectname
    if not os.path.exists(path):
        os.makedirs(path)
    click.echo("\033[1;32;40m Creating Project at %s.... " % path)
    click.echo("\033[1;32;40m Pulling Project Skeleton from Repo")
    Repo.clone_from(git_url, path)

    click.echo("\033[1;32;40m Setting up project")

    # Setting up the Config File
    config_path = path + "/config/config.ini"

    shutil.rmtree(path + "/.git")
    os.remove(path + "/.gitignore")

    config = ConfigParser()
    config.read(config_path)

    config.set('CONFIG', 'APP_NAME', projectname)
    hash_ = bcrypt.hashpw(str(projectname).encode('utf-8'), bcrypt.gensalt(12))
    config.set('CONFIG', 'APP_KEY', str(hash_))

    with open(config_path, 'w') as config_file:
        config.write(config_file)

    click.echo("\033[1;32;40m New Bast Project created at  %s " % path)


@main.command('create:migration', short_help="Create a migration file")
@click.argument('migration_file', required=1)
@click.option('--create', default=True, help="Create the table. OPTIONAL")
@click.option('--table', default=None, help="Name of the table to be created. OPTIONAL")
def migration_creatr(migration_file, create, table):
    """Name of the migration file"""
    bast_app = os.path.abspath('.') + '/config/config.ini'
    if not os.path.exists(bast_app):
        click.echo('ERROR: Ensure you are in a bast app to run the create:migration command')
        return

    migration = CreateMigration()
    if table is None:
        table = snake_case(migration_file)
    file = migration.create_file(snake_case(migration_file), table=table, create=create)
    click.echo('Migration file created at %s' % file)


@main.command('migration:run', short_help="Run Migration")
@click.option('--pretend', default=False, help="Simulates the Migration")
def migration_run(pretend):
    bast_app = os.path.abspath('.') + '/config/config.ini'
    if not os.path.exists(bast_app):
        click.echo('ERROR: Ensure you are in a bast app to run the migration:run command')
        return

    migration = Migration()
    migration.run_(pretend)
    click.echo('Migration Run successful')


@main.command('migration:rollback', short_help="Roll Back last Migration")
@click.option('--pretend', default=False, help="Simulates the Migration")
def migration_rollback(pretend):
    bast_app = os.path.abspath('.') + '/config/config.ini'
    if not os.path.exists(bast_app):
        click.echo('ERROR: Ensure you are in a bast app to run the migration:rollback command')
        return

    migration = Migration()
    count = migration.rollback_(pretend)
    click.echo('Roll Back Executed. %s migrations rolled back' % count)


@main.command('migration:reset', short_help="Reset Migration")
@click.option('--pretend', default=False, help="Simulate the Migration")
def migration_reset(pretend):
    bast_app = os.path.abspath('.') + '/config/config.ini'
    if not os.path.exists(bast_app):
        click.echo('ERROR: Ensure you are in a bast app to run the migration:rollback command')
        return

    migration = Migration()
    count = migration.reset_(pretend)
    click.echo('Migration Reset successful. %s migrations has been reset' % count)


@main.command('create:model', short_help="Create Model File")
@click.argument('model_file', required=1)
@click.option('--migration', default=True, help="Generate Migration File Also")
def model_creatr(model_file, migration):
    bast_app = os.path.abspath('.') + '/config/config.ini'
    if not os.path.exists(bast_app):
        click.echo('ERROR: Ensure you are in a bast app to run the create:model command')
        return

    filename = snake_case(model_file) + ".py"

    directory_path = os.path.abspath('.') + '/models/'
    if not os.path.exists(directory_path):
        os.makedirs(directory_path)

    path = os.path.abspath('.') + '/models/' + filename
    file_open = open(path, 'w+')
    compose = 'from bast import Models\n\nclass %s(Models):\n    __table__ = \'%s\'' \
              % (model_file, snake_case(model_file))
    # compose = 'from bast import Models\n\n\nclass %s(Models):\n    pass' % model_file
    file_open.write(compose)
    file_open.close()
    if migration:
        migrate = CreateMigration()
        migrate.create_file(name=snake_case(model_file), table=snake_case(model_file), create=True)
    click.echo('%s has been created at /models' % filename)


def snake_case(string_name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string_name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()



"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

import logging
import traceback

from tornado.web import RequestHandler

from .exception import BastException
from .json_ import Json as json_
from .view import TemplateRendering


class Controller(RequestHandler, TemplateRendering):

    def write_error(self, status_code, **kwargs):
        """
        Handle Exceptions from the server
        :param status_code:
        :param kwargs:
        :return:
        """
        reason = self._reason
        if self.settings.get("serve_traceback") and "exc_info" in kwargs:
            error = []
            for line in traceback.format_exception(*kwargs["exc_info"]):
                error.append(line)
        else:
            error = None
        self.write(html_error(status_code, reason, error))

    def view(self, template_name, kwargs=None):
        """
        This is for making some extra context variables available to
        the template
        """
        if kwargs is None:
            kwargs = dict()

        kwargs.update({
            'settings': self.settings,
            'STATIC_URL': self.settings.get('static_url_prefix', 'public/static/'),
            'request': self.request,
            'xsrf_token': self.xsrf_token,
            'xsrf_form_html': self.xsrf_form_html,
        })

        content = self.render_template(template_name, **kwargs)
        self.write(content)

    def data_received(self, chunk):
        pass

    def initialize(self, method):
        self.method = method

    def only(self, arguments):
        """
        returns the key, value pair of the arguments passed as a dict object
        Example usage
        data = self.only(['arg_name'])
        :param arguments:
        :return:
        """
        data = {}
        for i in arguments:
            data[i] = self.get_argument(i)
        return data

    def except_(self, arguments):
        """
        returns the arguments passed to the route except that set by user

        Example usage
        data = self.except_(['arg_name'])
        :param arguments:
        :return:
        """
        args = self.request.arguments
        data = {}
        for key, value in args.items():
            if key not in arguments:
                data[key] = self.get_argument(key)
        return data

    def json(self, data):
        """
        Encodes the dictionary being passed to JSON and sets the Header to application/json
        :param data:
        :return:
        """
        self.write(json_.encode(data))
        self.set_header('Content-type', 'application/json')

    def get(self, *args, **kwargs):
        # print('Args: ' + str(args))
        try:
            func = getattr(self, self.method)
            if func:
                func()
            else:
                raise BastException(404, "Not Found")
        except AttributeError as e:
            logging.error(str(e))
            raise BastException(500, "Controller Function not found")

    def post(self, *args, **kwargs):
        try:
            func = getattr(self, self.method)
            if func:
                func()
            else:
                raise BastException(404, "Not Found")
        except AttributeError as e:
            logging.error(str(e))
            raise BastException(500, "Controller Function not found")

    def put(self, *args, **kwargs):
        try:
            func = getattr(self, self.method)
            if func:
                func()
            else:
                raise BastException(404, "Not Found")
        except AttributeError as e:
            logging.error(str(e))
            raise BastException(500, "Controller Function not found")

    def delete(self, *args, **kwargs):
        try:
            func = getattr(self, self.method)
            if func:
                func()
            else:
                raise BastException(404, "Not Found")
        except AttributeError as e:
            logging.error(str(e))
            raise BastException(500, "Controller Function not found")

    # def validate(self, params):
    #     for i in range(0, len(params)):
    #         if self.get_argument(params[i]) is None:
    #             return False
    #     return True


def html_error(code, message, _traceback=None):
    """
    Render Error code as HTML for better viewing
    :param code:
    :param message:
    :param _traceback:
    :return:
    """
    message_ = """
            <!DOCTYPE html>
                <html lang="en">
                <head>
                    <meta charset="UTF-8">
                    <meta name="viewport" content="width=device-width, initial-scale=1.0">
                    <meta http-equiv="X-UA-Compatible" content="ie=edge">
                    <title>Arrrgghh!! Exception </title>
                    <link rel="stylesheet" href="style.css">
                    <style>
                        body {
                            margin: 0px auto;
                            padding: 0px;
                            font-family: 'Trebuchet MS', 'Lucida Sans Unicode', 'Lucida Grande', 'Lucida Sans', Arial, sans-serif;
                        }
                        
                        .h-text {
                            color: #993300;
                        }
                        
                        .container {
                            width: 70%;
                            margin: 40px auto;
                            padding: 20px;
                            box-shadow: 1px 1px 1px 1px #ccc;
                            background: #F8F6F8;
                        }
                        
                        .image {
                            width: 100px;
                            max-width: 100%;
                            opacity: 0.6;
                        }
                        
                        hr {
                            border-width: 1px;
                            border-color: #991100;
                        }
                        
                        .show-more{
                            padding-top: 10px;
                            padding-bottom: 10px;
                        }
                        .show-more a {
                            color: #993300;
                        }
                        
                        .panel-error {
                            background: #DDD;
                            padding: 10px;
                            color: #666;
                        }
                        
                        .panel-error-editor {
                            background: #DDD;
                            padding: 10px;
                            color: #666;
                        }
                        
                        .error-show {
                            color: #666;
                        }
                        
                        .class {
                            color: blue;
                            font-weight: bold;
                        }
                        
                        .method {
                            color: #666;
                        }
                        
                        .variable {
                            color: #111;
                        }

                    </style>
                    <script>
                        function show_more(e) {
                            e.preventDefault()
                            const el = document.querySelector('.panel-error-editor')
                            if( el.style.display == 'none') {
                                el.style.display = 'block'
                            } else {
                                el.style.display = 'none'
                            }
                        }
                    </script>
                </head>
                <body>
                    <div class="container">
                        <h1 class="h-text">""" + str(code) + " " + message + """
                             </h1>
                        <hr>
                        <p class="error-show"> An Error Occurred during Execution </p> """

    if _traceback is None:
        message_ += "</div></body>"
        return message_

    for i in range(0, len(_traceback)):
        if i is 0:
            message_ += "<div class='panel-error'> <h1 class='h-text'> %s </h1> " % (_traceback[0])
            continue
        message_ += "<p>%s</p>" % (_traceback[i])

    message_ += """
                        </div>
                        <div class="show-more">
                        </div>
                    </div>
                </body>
                </html>
    """

    return message_



"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""

import logging
import os

try:
    from configparser import ConfigParser
except ImportError:
    import ConfigParser

from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.options import define, options, parse_command_line
from tornado.web import Application, StaticFileHandler

__author__ = "Majiyagbe Oluwole"
__copyright__ = ""
__license__ = "M.I.T License"
__version__ = "1.0"
__status__ = "Under Development"


class Bast(Application):
    def __init__(self, route, **settings):
        """
         Bast Server Class. Runs on Tornado HTTP Server (http://www.tornadoweb.org/en/stable/)

        Constructor for the Bast Server. Takes an instance of the route as parameter.
        The Web handler with routes are handled here.

        Config files are also loaded from the config/config.ini folder.
        Appropriate configurations are loaded from the config file into the os environment for use
        :param route:
        """
        super().__init__(**settings)
        self.host = '127.0.0.1'
        self.port = 2000
        self.debug = True

        self.load_config()

        self.handler = route.all().url
        self.handler.append((r'/css/(.*)', StaticFileHandler, {"path": os.path.abspath(".") + "/public/static/css"}))
        self.handler.append((r'/script/(.*)', StaticFileHandler, {"path": os.path.abspath(".") + "/public/static/js"}))
        self.handler.append(
            (r'/images/(.*)', StaticFileHandler, {"path": os.path.abspath('.') + "/public/static/images"}))

    def run(self):
        """
        Function to Run the server. Server runs on host: 127.0.0.1 and port: 2000 by default. Debug is also set to false
        by default

        Can be overriden by using the config.ini file
        :return:
        """
        define("port", default=self.port, help="Run on given port", type=int)
        define("host", default=self.host, help="Run on given host", type=str)
        define("debug", default=self.debug, help="True for development", type=bool)

        parse_command_line()

        logging.info("Starting Bast Server....")
        logging.info("Bast Server Running on %s:%s" % (options.host, options.port))

        application = Application(self.handler, debug=options.debug)
        server = HTTPServer(application)
        server.listen(options.port, options.host)
        IOLoop.current().start()

    def load_config(self):
        """
        Function to load configuration details from the config.ini file into environment variables.
        :return:
        """
        config_path = os.path.abspath('.') + "/config/config.ini"
        if not os.path.exists(config_path):
            return

        config = ConfigParser()
        config.read(config_path)

        #   config section
        os.environ['APP_NAME'] = config['CONFIG']['APP_NAME']
        os.environ['APP_KEY'] = config['CONFIG']['APP_KEY']

        os.environ['DB_TYPE'] = config['DATABASE']['DB_TYPE']
        os.environ['DB_NAME'] = config['DATABASE']['DB_NAME']
        os.environ['DB_HOST'] = config['DATABASE']['DB_HOST']
        os.environ['DB_USER'] = config['DATABASE']['DB_USER']
        os.environ['DB_PASSWORD'] = config['DATABASE']['DB_PASSWORD']
        os.environ['DB_PREFIX'] = config['DATABASE']['DB_PREFIX']
        os.environ['ABS_PATH'] = os.path.abspath('.')

        self.host = config['CONFIG']['HOST']
        self.port = config['CONFIG']['PORT']
        self.debug = config['CONFIG']['DEBUG']

        os.environ['TEMPLATE_FOLDER'] = os.path.abspath('.') + "/public/templates"



"""
    Bast Web Framework
    (c) Majiyagbe Oluwole <oluwole564@gmail.com>

    For full copyright and license information, view the LICENSE distributed with the Source Code
"""
import importlib


class Route:
    """
    Route Class. Appends the URL, Controller Instance and Method to a list instance to be passed on to the server
    instance

    """

    def __init__(self):
        self.url = []
        self.middleware_list = []
        self.controller_location = 'controller'
        self.middleware_location = 'middleware'

    def middleware(self, *args):
        self.middleware_list = args
        return self

    def __run_middleware__(self):
        try:
            for func in self.middleware_list:
                middleware_func = importlib.import_module('{0}.'.format(self.middleware_location) + func)
                if hasattr(middleware_func, func):
                    class_name = getattr(middleware_func, func)
                    handle = getattr(class_name, 'handle')
                    return_value = handle(class_name)
                    print(return_value)
        except Exception as e:
            print("There is an Error in your Middleware ", e)

    def __return_controller__(self, controller):
        if isinstance(controller, str):
            ctr = controller.split('.')
            if ctr[0].startswith('/'):
                self.controller_location = '.'.join(ctr[0].replace('/', '').split('.')[0:-1])
        else:
            if controller is None:
                return None

            name = controller.__qualname__
            ctr = name.split('.')
            self.controller_location = controller.__module__

        get_controller = ctr[0].split('.')[-1]
        try:
            # Import the module
            print('{0}.'.format(self.controller_location) + get_controller)
            if isinstance(controller, str):
                controller_name = importlib.import_module('{0}.'.format(self.controller_location) + get_controller)
            else:
                controller_name = importlib.import_module('{0}'.format(self.controller_location))

            # Get the controller from the module
            controller_class = getattr(controller_name, get_controller)

            # Get the controller method from the class which in this case is the string
            controller_method = ctr[1]
            return controller_class, controller_method

        except Exception as e:
            print('\033[93mError in your routes/link.py!', e, '\033[0m')

    def get(self, url, controller):
        """
        Adds URL to the url list for GET request
        :param url:
        :param controller:
        :return:
        """
        controller_class, controller_method = self.__return_controller__(controller)
        self.url.append((url, controller_class, dict(method=controller_method)))
        return self

    def post(self, url, controller):
        """
        Adds URL to the url for the POST request
        :param url:
        :param controller:
        :return:
        """
        controller_class, controller_method = self.__return_controller__(controller)
        self.url.append((url, controller_class, dict(method=controller_method)))
        return self

    def put(self, url, controller):
        """
        Adds URL to list for PUT request
        :param url:
        :param controller:
        :return:
        """
        controller_class, controller_method = self.__return_controller__(controller)
        self.url.append((url, controller_class, dict(method=controller_method)))
        return self

    def delete(self, url, controller):
        """
        Adds URL for the delete request
        :param url:
        :param controller:
        :return:
        """
        controller_class, controller_method = self.__return_controller__(controller)
        self.url.append((url, controller_class, dict(method=controller_method)))
        return self

    def all(self):
        """
        Returns the list of URL. Used by Server to get the list of URLS
        :return:
        """
        self.__run_middleware__()
        return self
